/* mime64 */
/* MIME base64 encoder/decoder by Karl Hahn  hahn@lds.loral.com  3-Aug-94 */
/* Modified 30-Sep-94 by Karl Hahn hahn@lds.loral.com: handle multiple
   content */
/* Modified July 2007 by Eric Auer to reduce executable file size:
   avoid printf / fprintf / sprintf, avoid sscanf */
/* other ideas: avoid things that use malloc, avoid bufferd files */
/* Turbo C 2 compile example: tcc -lt -1- -A -Z -f87- -f- -mt -r -w
   -Ic:\tc\include -Lc:\tc\lib -M -v- -y- MIME64.C */
/* OpenWatcom: something like: wcc -0 -d0 -ms -os -s -wx -za -zl
     -zld -zls mime64.c ... ms2wlink /ma mime64.obj | wlink ...?
   WLINK input: system dos\nfile mime64.obj
     library c:\watcom\lib286\dos\clibs.lib option map */
/* Always required library functions: unlink rename */
/* Other library functions: either FILE or handle based file I/O */
/* #include <stdlib.h> - only for exit(n) ? */

#include <stdio.h>
#define USESTRING 0 /* set to 1 to use C lib strcpy and strcmpi */
#if USESTRING
#include <string.h>
#endif

#define USESTREAMS 0 /* set to 1 for f{open,close,eof,{put,get}{s,c}} */
#if USESTREAMS
/* ... */
#else
#include <io.h>
#include <fcntl.h>
#endif

#ifndef unlink /* huh? should be in stdio.h already */
int unlink (const char *path);
#endif

unsigned char alphabet[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
                  "abcdefghijklmnopqrstuvwxyz"
                  "0123456789+/";

enum TOKENTYPE { NONE, BLANKS, PUNCT, TAG, NAME, CONTENT };

struct TOKEN {
                char *text;
                int  length;
                int  index;
                enum TOKENTYPE type;
             };


#if USESTRING
#else
void strcpy( char *dest, char *src )
{
    while ( src[0]!='\0' )
    {
       dest[0] = src[0];
       dest++;
       src++;
    }
    dest[0] = '\0';
}

/* we only need a check for equality here, but have to ignore case */
int strcmpi( char *a, char *b )
{
   char aa, bb;
   while ( 1 )
   {
      aa = a[0];
      bb = b[0];
      if ( aa>='a' && aa<='z' ) aa = aa + 'A' - 'a';
      if ( bb>='a' && bb<='z' ) bb = bb + 'A' - 'a';
      if ( aa!=bb ) return -1;
      if ( aa=='\0' ) return 0;
      a++;
      b++;
   }
}
#endif

#if USESTREAMS
/* ... */
#else
int my_write( char *what, int handle )
{
   int size = 0;
   char *chk = what;
   while ( chk[0]!='\0' )
   {
      size++;
      chk++;
   }
   return write( handle, what, size );
}

/* no EOF or error checking needed in our case */
int my_getc( int handle )
{
   char gbuf[2];
   gbuf[0] = 0;
   read( handle, gbuf, 1 );
   return gbuf[0];
}

/* read until end of line, return null if end of file hit */
char * my_gets( char *dest, int size, int handle )
{
   int c, n;
   char * dptr = dest;
   n = 0;
   do
   {
      c = my_getc( handle );
      if ( eof( handle ) ) return 0;
      if ( c=='\r' || c=='\n' ) break;
      dptr[0] = c;
      dptr++;
      n++;
  } while ( n < size );
  dptr[0] = '\0';
  return dest;
}
#endif

/* we need no complex printf or fprintf, only our print and fputs */
void print( char *text )
{
#if USESTREAMS
   fputs( text, stdout );
#else
   my_write( text, 1 ); /* stdout */
#endif
}

int compare_token( struct TOKEN *token, char *text )
{
   int index=0;
   int count;
   int result;
   char blivit1, blivit2;

   count = token->length;

   if ( count > 0 )
   {
      result = 1;
   }
   else
   {
      result = 0;
   }

   while ( (count > 0) && ( result != 0 ) )
   {
      blivit1 = token->text[index++];
      if ( (blivit1 >= 'a' ) && (blivit1 <= 'z') )
      {
         blivit1 -= ' ';
      }

      blivit2 = *text++;
      if ( (blivit2 >= 'a' ) && (blivit2 <= 'z') )
      {
         blivit2 -= ' ';
      }

      if ( blivit1 != blivit2 )
      {
         result = 0;
      }

      count--;
   }

   return result;
}

int ispunct( char blivit )
{
   if ( ( ( blivit < '0' ) ||
          ( ( blivit > '9' ) && (blivit < 'A') ) ||
          ( blivit > 'Z' ) ) &&
        ( blivit != '-') && (blivit != '/') && (blivit != '.') )
   {
      return 1;
   }
   else
   {
      return 0;
   }
}

void acquire_token( char *line, enum TOKENTYPE type, struct TOKEN *token )
{
   int doneflag=0, startflag=1;
   int index;
   enum TOKENTYPE nextstate=NONE;
   char blivit;

   if (token->type == NONE)
   {
      token->index = 0;
      token->length = 0;
   }

   index = token->index + token->length;

   token->text = 0;

   while ( doneflag == 0 )
   {
      blivit = line[index];
      if ( (blivit >= 'a') && (blivit <= 'z') )
      {
         blivit -= ' ';
      }

      switch (token->type)
      {
         case NONE:
         if ( blivit == ' ')
         {
            index++;
            token->index++;
         }
         else
         {
            token->type = TAG;
            nextstate = TAG;
         }
         break;

         case BLANKS:
         if      ( blivit == ' ')
         {
            index++;
         }
         else if ( ispunct( blivit ) )
         {
            token->type = PUNCT;
            token->index = index;
         }
         else
         {
            token->type = nextstate;
            token->index = index;
         }
         break;

         case PUNCT:
         if      ( blivit <  ' ')
         {
            doneflag = 1;
            token->type = NONE;
            token->index = index;
            token->text = line + index;
            token->length = 0;
         }
         else if ( blivit == ' ' )
         {
            token->type = BLANKS;
            token->index = index;
            if      ( line[ token->index ] == ';' )
            {
               nextstate = NAME;
            }
            else if ( line[ token->index ] == '=' )
            {
               nextstate = CONTENT;
            }

         }
         else if ( ispunct( blivit ) )
         {
            index++;
         }
         else
         {
            if      ( line[ token->index ] == ';' )
            {
               nextstate = NAME;
            }
            else if ( line[ token->index ] == '=' )
            {
               nextstate = CONTENT;
            }

            token->type = nextstate;
            token->index = index;
         }
         break;

         case TAG:
         if ( ispunct( blivit ) )
         {
            token->length = index - token->index;
            token->text = line + token->index;
            nextstate = NAME;

            if ( ( ( type == TAG ) || ( type == NONE ) ) && !startflag)
            {
               doneflag = 1;
            }
            else if (blivit == ' ')
            {
               token->type = BLANKS;
               token->index = index;
            }
            else
            {
               token->type = PUNCT;
               token->index = index;
            }
         }
         else
         {
            index++;
         }
         break;

         case NAME:
         if ( ispunct( blivit ) )
         {
            token->length = index - token->index;
            token->text = line + token->index;
            nextstate = CONTENT;

            if ( ( ( type == NAME ) || ( type == NONE ) ) && !startflag )
            {
               doneflag = 1;
            }
            else if (blivit == ' ')
            {
               token->type = BLANKS;
               token->index = index;
            }
            else
            {
               token->type = PUNCT;
               token->index = index;
            }
         }
         else
         {
            index++;
         }
         break;

         case CONTENT:
         if ( ispunct( blivit ) )
         {
            token->length = index - token->index;
            token->text = line + token->index;
            nextstate = NAME;

            if ( ( ( type == CONTENT ) || ( type == NONE ) ) && !startflag )
            {
               doneflag = 1;
            }
            else if (blivit == ' ')
            {
               token->type = BLANKS;
               token->index = index;
            }
            else
            {
               token->type = PUNCT;
               token->index = index;
            }
         }
         else
         {
            index++;
         }
         break;
      }
      startflag = 0;
   }
}

#if 0
#if USESTREAMS
void fputch( char blivit, FILE *f )
{
/*   if (blivit == '\n') fputc( '\r', f );*/
   fputc( blivit, f );
}
#else
/* ... */
#endif
#endif

int classify_args( int narg,
                   char *rawargs[], char *fileargs[], char *optargs[] )
{
   int index, jndex, kndex;
   char *argptr;

   for ( index = 0, jndex = 0, kndex = 0; index < narg; index++ )
   {
      argptr = rawargs[index];
      if (*argptr == '-')
      {
         argptr++;
         optargs[kndex++] = argptr;
      }
      else
      {
         fileargs[jndex++] = argptr;
      }
   }

   return kndex;
}

int cvt_ascii( unsigned char alpha )
{
   if      ( (alpha >= 'A') && (alpha <= 'Z') ) return (int)(alpha - 'A');
   else if ( (alpha >= 'a') && (alpha <= 'z') )
        return 26 + (int)(alpha - 'a');
   else if ( (alpha >= '0') && (alpha <= '9' ) )
        return 52 + (int)(alpha - '0');
   else if ( alpha == '+' ) return 62;
   else if ( alpha == '/' ) return 63;
   else if ( alpha == '=' ) return -2;
   else                     return -1;
}

char *fileargs[64], *optargs[64];

struct STATE64 {
                  unsigned long int accum;
                  int               shift;
               };


int main( int nargs, char *cargs[] )
{
   int n_options, n_files, index, shift, save_shift;
   /* removed unused: int jndex, FILE *dummy, char *cptr, char *altptr */
   enum { ENCODE, DECODE } whattodo = DECODE;
   int help_flag = 0, replace_flag = 0, quit = 0;
#if USESTREAMS
   FILE *fin, *fout;
#else
   int infi, outfi;
   char xbuf[2];
#endif
   unsigned char blivit;
   unsigned long accum, value;
   char buf[80], dumname[80];
   char *siptr, *soptr;
   int decode_state;
   struct TOKEN token;
   int firsttime = 1;
   int skipflag = 0;
   int printmsg = 1;

   n_options = classify_args( nargs, cargs, fileargs, optargs );

   n_files = nargs - n_options;

   if ( n_files < 2 ) help_flag = 1;

   for ( index = 0; index < n_options; index++ )
   {
      if ( ( optargs[index][0] == 'e' ) ||
           ( optargs[index][0] == 'E' ) ) whattodo = ENCODE;
      if ( optargs[index][0] == '?' ) help_flag = 1;
   }

   if ( help_flag )
   {
      print( "mime64 infile [outfile] [-option] [-option] etc.\r\n\r\n"
              "convert between binary and MIME BASE64 format\r\n\r\n"
              "        -e       MIME base64 encode (default is decode)\r\n"
              "        -?       display help message\r\n\r\n"
              "if no outfile given, output file replaces infile\r\n" );
   }

   if ( n_files < 2 ) return 0; /* exit(0); */

   if ( whattodo == DECODE )
   {
#if USESTREAMS
      fin = fopen( fileargs[1], "r" );
#else
      infi = open( fileargs[1], O_RDONLY );
#endif
   }
   else
   {
#if USESTREAMS
      fin = fopen( fileargs[1], "rb" );
#else
      infi = open( fileargs[1], O_RDONLY | O_BINARY);
#endif
   }

#if USESTREAMS
   if ( fin == 0 )
#else
   if ( infi == -1 )
#endif
   {
      print( fileargs[1] );
      print( " file not found\r\n" );
      return -1; /* exit(-1); */
   }

   if ( n_files > 2 )
   {
      if ( whattodo == DECODE )
      {
         /* sprintf( dumname, "%s", fileargs[2] ); */
         strcpy( dumname, fileargs[2] );
      }
      else
      {
#if USESTREAMS
         fout = fopen( fileargs[2], "w" );
#else
         outfi = open( fileargs[2], O_WRONLY );
#endif
      }

#if USESTREAMS
      if ( fout == 0 )
#else
      if ( outfi == -1 )
#endif
      {
         print( "Couldn't open " );
         print( fileargs[2] );
         print( " for output\r\n" );
      }
   }
   else
   {
      if ( whattodo == DECODE )
      {
         /* sprintf( dumname, "%s", fileargs[1] ); */
         strcpy( dumname, fileargs[1] );
      }
      else
      {
#if USESTREAMS
         fout = fopen( "$$$$$$$$.$$$", "w" );
#else
         outfi = open( "$$$$$$$$.$$$", O_WRONLY );
#endif
      }

      replace_flag = 1;
   }


do {
   quit = 0;
   printmsg = 1;

   if ( whattodo == DECODE )
   {
      shift = 0;
      accum = 0;
      decode_state = 0;

#if USESTREAMS
      while ( ( !feof( fin ) ) && (quit == 0) )
      {
         fgets( buf, 80, fin );
         if ( feof( fin ) )
#else
      while ( ( !eof( infi ) ) && (quit == 0) )
      {
         my_gets( buf, 80, infi );
         if ( eof( infi ) )
#endif
         {
            if ( dumname[0] != '\0' )
            {
               print( "Unexpected end of file encountered in " );
               print( dumname );
               print( "\r\n" );
            }

            return 0; /* exit(0); */
         }

         if ( decode_state == 0 )
         {
            for ( index = 0;
                  (buf[index] != '\n') && (buf[index] != '\0') &&
                  (decode_state >= 0);
                  index++ )
            {
               if ( ( (buf[index] >= 'A') && (buf[index] <= 'Z') ) ||
                    ( (buf[index] >= 'a') && (buf[index] <= 'z') ) ||
                    ( (buf[index] >= '0') && (buf[index] <= '9') ) ||
                    (buf[index] == '+') ||
                    (buf[index] == '/') ||
                    (buf[index] == '=') )
               {
                  decode_state = 1;
               }
               else
               {
                  decode_state = -2;
               }
            }

            if ( decode_state <= 0 )
            {
               decode_state = 0;
               token.type = NONE;

               acquire_token( buf, TAG, &token );
               if      ( compare_token( &token, "Content-Type") )
               {
                  do
                  {
                     acquire_token( buf, NAME, &token );
                     if ( compare_token( &token, "name" ) )
                     {
                        acquire_token( buf, CONTENT, &token );

                        if ( ( replace_flag ) ||
                             ( firsttime == 0 ) )
                        {
                           /* read token.text to dumname */
                           /* sscanf( token.text, "%s", dumname ); */
                           siptr = token.text;
                           soptr = dumname;
                           while ( siptr[0]!=' ' && siptr[0]!='\r' &&
                              siptr[0]!='\n' && siptr[0]!='\t' )
                           {
                              soptr[0] = siptr[0];
                              soptr++;
                              siptr++;
                              if (soptr==(dumname+60)) soptr--;
                           }
                           soptr[0] = '\0';
                           replace_flag = 0;
                        }
                     }
                  } while ( token.type != NONE );
               }
               else if ( compare_token( &token, "Content-transfer-encoding" ) )
               {
                  skipflag = 1;

                  do
                  {
                     acquire_token( buf, NAME, &token );
                     if ( compare_token( &token, "base64" ) )
                     {
                        skipflag = 0;
                     }
                  } while ( token.type != NONE );
               }
               continue;
            }
            else if ( skipflag != 0 )
            {
               continue;
            }
         }

         if ( printmsg )
         {
            if ( skipflag )
            {
               print( "Section " );
               print( dumname );
               print( " not MIME base64\r\n" );
            }
            else
            {
               print( "Creating " );
               print( dumname );
               print( "\r\n" );
               if ( strcmpi( dumname, fileargs[1] ) == 0 )
               {
                  replace_flag = 1;
               }

               if ( replace_flag )
               {
#if USESTREAMS
                  fout = fopen( "$$$$$$$$.$$$", "wb" );
#else
                  outfi = open( "$$$$$$$$.$$$", O_WRONLY | O_BINARY );
#endif
               }
               else
               {
#if USESTREAMS
                  fout = fopen( dumname, "wb" );
#else
                  outfi = open( dumname, O_WRONLY | O_BINARY );
#endif
               }

#if USESTREAMS
               if ( fout == 0 )
#else
               if ( outfi == -1 )
#endif
               {
                  print( "Couldn't open " );
                  print( dumname );
                  print( " for output\r\n" );
               }
            }

            printmsg = 0;
         }

#if USESTREAMS
         if ( fout == 0 )
#else
         if ( outfi == -1 )
#endif
         {
            print( "No filename given for subsequent section\r\n" );
            return -1; /* exit(-1); */
         }

#if USESTREAMS
         if ( feof(fin) )
#else
         if ( eof(infi) )
#endif
         {
            quit = 1;
         }

         if ( quit != 0 )
         {
            buf[0] = '\0';
         }

         for ( index = 0; (buf[index] != '\n') && (buf[index] != '\0'); index++)
         {
            value = cvt_ascii( buf[index] );

            if ( value < 64 )
            {
               accum <<= 6;
               shift += 6;
               accum |= value;
               if ( shift >= 8 )
               {
                  shift -= 8;
                  value = accum >> shift;
                  blivit = (unsigned char)value & 0xFFl;
#if USESTREAMS
                  fputc( blivit, fout );
#else
                  xbuf[0] = blivit;
                  write( outfi, xbuf, 1 );
#endif
               }
            }
            else
            {
               quit = 1;
               break;
            }
         }
      }
   }
   else
   {
      /* fprintf ( fout,
       "Content-Type: text/plain; charset=US-ASCII; name=%s\r\n"
       "Content-transfer-encoding: base64\r\n\r\n", fileargs[1] ); */
#if USESTREAMS
      fputs ( "Content-Type: text/plain; charset=US-ASCII; name=", fout );
      fputs ( fileargs[1], fout );
      fputs ( "\r\nContent-transfer-encoding: base64\r\n\r\n", fout );
#else
      my_write( "Content-Type: text/plain; charset=US-ASCII; name=", outfi );
      my_write( fileargs[1], outfi );
      my_write( "\r\nContent-transfer-encoding: base64\r\n\r\n", outfi );
#endif

      shift = 0;
      accum = 0;
      index = 0;
#if USESTREAMS
      while ( ( !feof( fin ) ) || (shift != 0) )
      {
         if ( ( !feof( fin ) ) && ( quit == 0 ) )
         {
            blivit = fgetc( fin );

            if ( feof( fin ) )
#else
      while ( ( !eof( infi ) ) || (shift != 0) )
      {
         if ( ( !eof( infi ) ) && ( quit == 0 ) )
         {
            blivit = my_getc( infi );

            if ( eof( infi ) )
#endif
            {
               quit = 1;
               save_shift = shift;
               blivit = 0;
            }
         }
         else
         {
            quit = 1;
            save_shift = shift;
            blivit = 0;
         }

         if ( (quit == 0) || (shift != 0) )
         {
            value = (unsigned long)blivit;
            accum <<= 8;
            shift += 8;
            accum |= value;
         } /* ENDIF */

         while ( shift >= 6 )
         {
            shift -= 6;
            value = (accum >> shift) & 0x3Fl;
            blivit = alphabet[(unsigned char)value & 0x3F];

            buf[index++] = blivit;
            if ( index >= 60 )
            {
               buf[index] = '\0';
#if USESTREAMS
               fputs( buf, fout );
               fputs( "\r\n", fout );
#else
               my_write( buf, outfi );
               my_write( "\r\n", outfi );
#endif
               index = 0;
            }

            if ( quit != 0 )
            {
               shift = 0;
            }
         }
      }

      if      ( save_shift == 2 )
      {
         buf[index++] = '=';
         if ( index >= 60 )
         {
            buf[index] = '\0';
#if USESTREAMS
            fputs( buf, fout );
            fputs( "\r\n", fout );
#else
            my_write( buf, outfi );
            my_write( "\r\n", outfi );
#endif
            index = 0;
         }

         buf[index++] = '=';
         if ( index >= 60 )
         {
            buf[index] = '\0';
#if USESTREAMS
            fputs( buf, fout );
            fputs( "\r\n", fout );
#else
            my_write( buf, outfi );
            my_write( "\r\n", outfi );
#endif
            index = 0;
         }
      }
      else if ( save_shift == 4 )
      {
         buf[index++] = '=';
         if ( index >= 60 )
         {
            buf[index] = '\0';
#if USESTREAMS
            fputs( buf, fout );
            fputs( "\r\n", fout );
#else
            my_write( buf, outfi );
            my_write( "\r\n", outfi );
#endif
            index = 0;
         }
      }

      if ( index != 0 )
      {
         buf[index] = '\0';
#if USESTREAMS
         fputs( buf, fout );
         fputs( "\r\n", fout );
#else
         my_write( buf, outfi );
         my_write( "\r\n", outfi );
#endif
      }
   }

#if USESTREAMS
   fclose( fout );
#else
   close( outfi );
#endif

   if ( replace_flag )
   {
      if ( ( whattodo == DECODE ) && ( decode_state <= 0 ) )
      {
         unlink( "$$$$$$$$.$$$" ); /* remove(...) */
         print( "No MIME base64 lines found in " );
         print( fileargs[1] );
         print( "\r\n" );
      }
      else
      {
         unlink( fileargs[1] ); /* remove(...) */
         rename( "$$$$$$$$.$$$", fileargs[1] );
      }
   }
   else
   {
      if ( ( whattodo == DECODE ) && ( decode_state <= 0 ) )
      {
         unlink( fileargs[2] ); /* remove(...) */
         print( "No MIME base64 lines found in " );
         print( fileargs[1] );
         print( "\r\n" );
      }
   }

#if USESTREAMS
   fout = 0;
#else
   outfi = -1;
#endif
   firsttime = 0;
   dumname[0] = '\0';
#if USESTREAMS
} while ( !feof( fin ) );

fclose( fin );
#else
} while ( !eof( infi ) );

close( infi );
#endif
return 0;
}
